//
// Created by Tom on 08/03/2023.
//

#ifndef TYPON_LIST_HPP
#define TYPON_LIST_HPP

#include <algorithm>
#include <ostream>
#include <vector>
//#include <nanobind/stl/detail/nb_list.h>
#include <pybind11/stl.h>

namespace typon {

template <typename T> class PyList {
public:
  using value_type = T;
  PyList(std::shared_ptr<std::vector<T>> &&v) : _v(std::move(v)) {}
  PyList(std::vector<T> &&v)
      : _v(std::move(std::make_shared<std::vector<T>>(std::move(v)))) {}

  PyList(std::initializer_list<T> &&v)
      : _v(std::make_shared<std::vector<T>>(std::move(v))) {}

  PyList() : _v(std::make_shared<std::vector<T>>()) {}

  auto begin() const { return _v->begin(); }

  auto end() const { return _v->end(); }

  METHOD(auto, append, (Self self, const T &x), {
    self._v->push_back(x);
  })

  METHOD(auto, py_contains, (Self self, const T &x), {
    return std::find(self.begin(), self.end(), x) != self.end();
  })

  auto size() const { return _v->size(); }

  void push_back(const T &value) { _v->push_back(value); }

  void clear() { _v->clear(); }

  /*operator std::vector<T>() const {
    return std::vector<T>(this->begin(), this->end());
  }

  operator std::vector<T> &() {
    return *reinterpret_cast<std::vector<T> *>(this);
  }*/

  constexpr const T &operator[](size_t i) const { return _v->operator[](i); }
  constexpr T &operator[](size_t i) { return _v->operator[](i); }

  size_t py_len() const { return _v->size(); }

  void py_repr(std::ostream &s) const {
    s << '[';
    if (_v->size() > 0) {
      repr_to(_v->operator[](0), s);
      for (size_t i = 1; i < _v->size(); i++) {
        s << ", ";
        repr_to(_v->operator[](i), s);
      }
    }
    s << ']';
  }

  void py_print(std::ostream &s) const { py_repr(s); }

  PyList<T> operator+(const PyList<T> &other) const {
    std::vector<T> v;
    v.reserve(_v->size() + other._v->size());
    v.insert(v.end(), _v->begin(), _v->end());
    v.insert(v.end(), other._v->begin(), other._v->end());
    return PyList<T>(std::move(v));
  }

  PyList<T> operator*(size_t n) const {
    PyList<T> v{};
    v._v->reserve(this->_v->size() * n);
    for (size_t i = 0; i < n; i++) {
      v._v->insert(v._v->end(), this->_v->begin(), this->_v->end());
    }
    return v;
  }

private:
  std::shared_ptr<std::vector<T>> _v;
};





}

template <typename T> typon::PyList<T> list(std::initializer_list<T> &&v) {
  return typon::PyList<T>(std::move(v));
}

namespace PYBIND11_NAMESPACE {
namespace detail {
template <typename Type>
struct type_caster<typon::PyList<Type>> : list_caster<typon::PyList<Type>, Type> {};
}}

/*NAMESPACE_BEGIN(NB_NAMESPACE)
NAMESPACE_BEGIN(detail)

template <typename Type> struct type_caster<PyList<Type>>
    : list_caster<PyList<Type>, Type> { };

NAMESPACE_END(detail)
NAMESPACE_END(NB_NAMESPACE)
*/
#endif // TYPON_LIST_HPP
