# coding: utf-8
import ast
from dataclasses import dataclass, field

from transpiler.phases.typing import FunctionType, ScopeKind, VarDecl, VarKind, TY_NONE
from transpiler.phases.typing.common import ScoperVisitor
from transpiler.phases.typing.types import PromiseKind, Promise, BaseType


@dataclass
class ScoperClassVisitor(ScoperVisitor):
    fdecls: list[(ast.FunctionDef, BaseType)] = field(default_factory=list)

    def visit_AnnAssign(self, node: ast.AnnAssign):
        assert node.value is None, "Class field should not have a value"
        assert node.simple == 1, "Class field should be simple (identifier, not parenthesized)"
        assert isinstance(node.target, ast.Name)
        self.scope.obj_type.members[node.target.id] = self.visit_annotation(node.annotation)

    def visit_FunctionDef(self, node: ast.FunctionDef):
        from transpiler.phases.typing.block import ScoperBlockVisitor
        # TODO: maybe merge this code with ScoperBlockVisitor.visit_FunctionDef
        argtypes = [self.visit_annotation(arg.annotation) for arg in node.args.args]
        argtypes[0].unify(self.scope.obj_type)  # self parameter
        rtype = self.visit_annotation(node.returns)
        inner_rtype = rtype
        if node.name != "__init__":
            rtype = Promise(rtype, PromiseKind.TASK)
        ftype = FunctionType(argtypes, rtype)
        self.scope.obj_type.methods[node.name] = ftype
        scope = self.scope.child(ScopeKind.FUNCTION)
        scope.obj_type = ftype
        scope.function = scope
        node.inner_scope = scope
        node.type = ftype
        for arg, ty in zip(node.args.args, argtypes):
            scope.vars[arg.arg] = VarDecl(VarKind.LOCAL, ty)
        res = (node, inner_rtype)
        self.fdecls.append(res)
        return res
